
/********************************************************************/
/*     classtree.c                                                  */
/*     -----------                                                  */
/*   Functions for processing Classtrees                            */
/********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "classtree.h"
#include "defns.i"  /* for Boolean type */


/*********************************************************************
 * NewClassTreeElem
 ********************************************************************/

struct Classtree *
NewClassTreeElem(char *n,struct Classtree *sib,struct Classtree *par,
		 struct Classtree *chd)
{
  struct Classtree* elem;
  if((elem = (struct Classtree *) malloc(sizeof (struct Classtree))) == NULL)
    {
      printf("Memory allocation error in allocating struct ClassTree in newClassTreeElem\n");
      exit(1);
    }
  else {
    if(strlen(n) < NAMELENGTH)
      { strcpy(elem->name, n); }
    else 
      { printf("newClassTreeElem: Error: name too long: <%s>\n",n); exit(1); }
    
    elem->sibling = sib;
    elem->parent  = par;
    elem->child   = chd;
    
  }
  return elem;
}

/**********************************************************************
 * Sets 
 *********************************************************************/

struct Classtree * SetSibling(struct Classtree * tree,struct Classtree * sibling)
{
  tree->sibling = sibling;
  return tree;
}

struct Classtree * SetParent(struct Classtree * tree,struct Classtree * parent)
{
  tree->parent = parent;
  return tree;
}

struct Classtree * SetChild(struct Classtree * tree,struct Classtree * child)
{
  tree->child = child;
  return tree;
}



/*********************************************************************
 * PrintClassTree
 *********************************************************************/

void PrintClassTree(const struct Classtree * tree, int indent)
{ 
  int i;

  if(tree)
    {
      for(i=1;i<=indent;i++)
	{
	  printf(" ");
	}
      printf("%s %d\n",tree->name, tree->arraypos);
      PrintClassTree(tree->child, indent+1);
      PrintClassTree(tree->sibling, indent);
    }

}

/****************************************************************/

int SizeOfTree(const struct Classtree * tree)
{
  int i=0;
  if(tree)
    {
      i++;
      if(tree->child) { i = i + SizeOfTree(tree->child); }
      if(tree->sibling) { i = i + SizeOfTree(tree->sibling); }
    }
  return i;
}

/****************************************************************/

/* List passed in must have enough space allocated already. 
 * Space is not allocated here!
 */

Boolean GetListOfChildren(struct Classtree * tree, struct Classtree **list)
{
  int i=0;
  struct Classtree *child;
  if(tree && tree->child)
    {
      child = tree->child;
      list[i] = child;
      i++;
      while(child->sibling)
      {
	child = child->sibling;
	list[i] = child;
	i++;
      }
      list[i] = NULL;
      return true;
    }
  else return false;
}

Boolean GetListOfChildrenIfExist(struct Classtree * tree, struct Classtree **list, ClassList existingclasses)
{
  int i=0;
  struct Classtree *child;
  if(tree && tree->child)
    {
      child = tree->child;
      /*      printf("%d ",child->arraypos);*/
      if(existingclasses[child->arraypos]) {
	list[i] = child;
	i++;
	/*	printf("* "); */
      }
      while(child->sibling)
      {
	child = child->sibling;
	/*	printf("%d ",child->arraypos);*/
	if(existingclasses[child->arraypos]) {
	  list[i] = child;
	  i++;
	  /*	  printf("* "); */
	}
      }
      /*      printf(" <%d> \n",i);*/
      list[i] = NULL;
      if(i) { return true; } else { return false; } 
    }
  else return false;
}


/*********************************************************************/
/* Queue for the search through tree                                 */
/*********************************************************************/

/* We add 1 to the size when creating it, to hold a NULL at the end */
struct Classtree** NewQueueOfClassTreeElems(int number)
{
  struct Classtree **p;
  if((p = calloc(number+1 , sizeof(struct Classtree*))) == NULL)
    {
      printf("NewQueueOfClassTreeElems: Error in allocating memory\n"); 
      exit(1);
    }
  else {p[0] = NULL; return p;}
}

struct Classtree** NewQueueOfClassTreeElems2(int number)
{
  struct Classtree **p;
  if((p = calloc(number+1 , sizeof(struct Classtree*))) == NULL)
    {
      printf("NewQueueOfClassTreeElems: Error in allocating memory\n"); 
      exit(1);
    }
  else {p[0] = NULL; return p;}
}

/****************************************************************/

void FreeQueueOfClassTreeElems( struct Classtree** queue )
{
  free(queue);
}

/****************************************************************/

/* User needs to check space okay beforehand */
void AddToQueue(struct Classtree** queue,struct Classtree** listofelems)
{
  int i,j;
  i=j=0;
  /* find end of queue */
  while(queue[i]) {i++;}
  /* copy in items */
  while(listofelems[j])
    {
      queue[i+j] = listofelems[j];
      j++;
    }
  queue[i+j] = NULL; 
}

/****************************************************************/

void AddOneToQueue(struct Classtree ** queue, struct Classtree *item)
{
  int i=0;
  /* find end of queue */
  while(queue[i]) {i++;}
  /* add item */
  queue[i] = item;
  queue[i+1] = NULL;
}

/****************************************************************/

struct Classtree * PopFromQueue(struct Classtree** queue)
{
  struct Classtree* first;
  int i = 1;
  if(queue[0])
    {
      first = queue[0];
      while(queue[i])
	{
	  queue[i-1] = queue[i];
	  i++;
	}
      queue[i-1]=NULL;
      return first;
    }
  else {return NULL;}
}

/****************************************************************/

Boolean EmptyQueue( struct Classtree** queue)
{
  if(queue[0]) { return false; } else { return true; }
}

/****************************************************************/

void PrintQueue( struct Classtree** queue)
{
  struct Classtree** p;
  static int i = 0;
  if (i > 100) {return;}
  p = queue;
  printf("Queue:");
  while(*p)
    {
      printf("<%s> ",(*p)->name);
      p++;
    }
  printf("\n");
  i++;
}

/******************************************************************/
/* Counting and finding things in the tree                        */
/******************************************************************/

int NumberClassTree(struct Classtree* tree,int from)
{
  int sofar = from;

  if(tree->child) {
    sofar = NumberClassTree(tree->child,sofar);
  }
  if(tree->sibling) {
    sofar = NumberClassTree(tree->sibling,sofar);
  }
  tree->arraypos = sofar;
  return (sofar+1);
}

/****************************************************************/

int NumberDescendantsClassTree(struct Classtree* tree)
{
  int sibs = 0, child = 0;
  if(tree->sibling) { 
    sibs = NumberDescendantsClassTree(tree->sibling); 
  } 
  if(tree->child) {
    child = NumberDescendantsClassTree(tree->child);
  }
  tree->numdescendants = child; 
  return sibs + child + 1;
}

/****************************************************************/

void ConstructInvertedIndex(struct Classtree* tree,
			    struct Classtree**InvertedNodeIndex)
{
  if(tree)
    {
      if(tree->sibling) {
	ConstructInvertedIndex( tree->sibling, InvertedNodeIndex );
      }
      if(tree->child) {
	ConstructInvertedIndex( tree->child, InvertedNodeIndex );
      }
      InvertedNodeIndex[ tree->arraypos ] = tree;
    }
}

/****************************************************************/

int GetIndexFromNode(struct Classtree* tree)
{
  if(!tree) {printf("GetIndexFromNode: null tree\n"); exit(1);}
  else {return tree->arraypos;}
}

/****************************************************************/

struct Classtree * GetNodeFromName(struct Classtree* tree, char* name)
{
  struct Classtree * node=NULL;
  if(!tree) {return tree;}
  else {
    if(strcmp(tree->name,name) == 0) {return tree;}
    else if(tree->child)
      { node = GetNodeFromName(tree->child,name); }
    if(!node && tree->sibling)
      { node = GetNodeFromName(tree->sibling,name); }
    return node;
  }
}

/****************************************************************/
/*
struct Classtree* GetNodeFromIndex2(int index,struct Classtree* tree)
{
  if(tree->arraypos < index)
    { printf("GetNodeFromIndex: index not found: %d\n",index); exit(0); }
  else if(index == tree->arraypos)
    {
      return tree;
    }
  else
    {
      if(tree->child && tree->child->arraypos >= index)
	{
	  return GetNodeFromIndex2(index,tree->child);
	}
      else {
	if(tree->sibling)
	  {
	    return GetNodeFromIndex2(index,tree->sibling);
	  }
	else { printf("GetNodeFromIndex: index not found: %d\n",index); exit(0); }
      }
    }
}
*/

struct Classtree* GetNodeFromIndex(int index,struct Classtree** InvertedNodeIndex)
{
  /*
  if(index > MaxClass) {
    fprintf(stderr,"index %d > MaxClass %d in GetNodeFromIndex\n"); exit(1); 
  }
  */
  return InvertedNodeIndex[index];
}

/****************************************************************/

int GetTreeSize(struct Classtree * tree)
{
  return tree->arraypos;
}

/****************************************************************/
/*
int GetNumDescFromIndex2( int index, struct Classtree* tree)
{
  struct Classtree * t = GetNodeFromIndex2(index,tree);
  return t->numdescendants;
}
*/

/****************************************************************/

int GetNumDescFromIndex( int index, struct Classtree** InvertedNodeIndex)
{
  struct Classtree * t = GetNodeFromIndex(index,InvertedNodeIndex);
  return t->numdescendants;
}


/******************************************************************/
/* ClassLists and classmasks to represent viewing the different   */
/* levels in hierarchy                                            */
/******************************************************************/

ClassList NewClassList(int size)
{
  ClassList p; 
  int i;

  if((p = calloc(size , sizeof(Boolean))) == NULL)
    {
      printf("NewClassList: Error in allocating memory\n"); 
      exit(1);
    }
  else {
    for(i=0; i<size; i++)
      p[i]=false;
    return p;
  }
}

/****************************************************************/

void FreeClassList( ClassList classlist)
{
  free(classlist);
}

/****************************************************************/

void CopyClassList(ClassList classlist1, ClassList classlist2, int length)
{
  int i;
  for(i=0; i<length; i++)
    {
      classlist1[i] = classlist2[i];
    }
}

/****************************************************************/

void CopyClassMaskedList( ClassList classlist1, ClassList classlist2, 
                          ClassList mask, int length)
{
  int i;
  for(i=0; i<length; i++)
    {
      classlist1[i] = classlist2[i] && mask[i];
    }
}

/****************************************************************/

void ApplyMask(ClassList classlist,ClassList mask,ClassList resultlist, int length)
{
  int i;
  for(i=0; i<length; i++)
    {
      resultlist[i] = classlist[i] && mask[i];
    }
}

/****************************************************************/

void SetBits(ClassList mask,struct Classtree** ontrees, int masklength)
{
  int i;
  for(i=0; i<masklength; i++) { mask[i]=0; }
  i=0;
  while(ontrees[i])
    {
      if(ontrees[i]->arraypos >= masklength) {
	printf("SetBits: arraypos %d bigger than mask size %d\n",
	       ontrees[i]->arraypos,masklength); 
	exit(1); 
      }
      else { mask[ ontrees[i]->arraypos ] = true; i++;}
    }
}

/****************************************************************/

void SetBitsOffAndOn(ClassList mask,struct Classtree* offtree, struct Classtree** ontrees, int masklength)
{
  int i=0;
  if(offtree && offtree->arraypos < masklength) 
    { mask[ offtree->arraypos ] = false;}
  while(ontrees[i])
    {
      if(ontrees[i]->arraypos >= masklength) 
	{ printf("SetBits: arraypos bigger than mask size\n"); exit(1); }
      else { mask[ ontrees[i]->arraypos ] = true; i++; }
    }
}

/****************************************************************/

Boolean AreEqual(ClassList list1, ClassList list2, ClassList mask, int length)
{
  int i = 0;
  Boolean areequal = true;
  while(areequal && i < length)
    {
      if( (list1[i] && mask[i])  !=  (list2[i] && mask[i]) )
	{ areequal = false; }
      i++;
    }
  return areequal;
}

/****************************************************************/
/*
void PrintClassList(ClassList list, struct Classtree* tree, int length)
{
  int i;
  struct Classtree* node;
  for(i=0; i <length; i++)
    {
      node=GetNodeFromIndex2(i,tree);
      printf("%d:%d %s\n",i,list[i],node->name);
    }
  printf("\n");
}

void PrintMaskedClassList(ClassList list,ClassList mask, struct Classtree* tree, int length)
{
  int i;
  struct Classtree* node;
  for(i=0; i <length; i++)
    {
      if(mask[i] && list[i])
	{
	  node=GetNodeFromIndex2(i,tree);
	  printf("%d:%d %s\n",i,list[i],node->name);
	}
    }
  printf("\n");
}
*/
/****************************************************************/

int NumberOfClassListClasses(ClassList pattern,ClassList mask, int length)
{
  int i,j=0;
  for(i=0; i <length; i++)
    {
      if(pattern[i] && mask[i]) { j++; }
    }
  return j;
}

/****************************************************************/
