/********************************************************************/
/*     classtree.h                                                  */
/*     -----------                                                  */
/*  Prototypes for functions in classtree.c and classtreereader.c   */
/********************************************************************/

#include "types.i"

#define NAMELENGTH 200

struct Classtree
{
  char name[NAMELENGTH];
  struct Classtree * sibling;
  struct Classtree * parent;
  struct Classtree * child;
  int arraypos;
  int numdescendants;
};

/*****************************************************************/

struct Classtree * NewClassTreeElem( char *n, struct Classtree *sib,
                                     struct Classtree *par,
				     struct Classtree *chd );

struct Classtree * SetSibling( struct Classtree *tree,struct Classtree *sibling);
struct Classtree * SetParent(  struct Classtree *tree,struct Classtree *parent);
struct Classtree * SetChild(   struct Classtree *tree,struct Classtree *child);
void               PrintClassTree( const struct Classtree * tree, int indent );
Boolean            GetListOfChildren( struct Classtree * tree,  
                                      struct Classtree **list);
Boolean            GetListOfChildrenIfExist(struct Classtree * tree, 
					    struct Classtree **list, 
					    ClassList existingclasses);
/* Boolean            IsChildOf( int i, int j, struct classtree *tree); */


/*****************************************************************/

struct Classtree** NewQueueOfClassTreeElems(int number);
struct Classtree** NewQueueOfClassTreeElems2(int number);

void               AddToQueue( struct Classtree** queue,
			       struct Classtree** listofelems);
void               AddOneToQueue( struct Classtree ** queue, 
				  struct Classtree *item);
struct Classtree * PopFromQueue( struct Classtree** queue);
Boolean            EmptyQueue( struct Classtree** queue);
void               FreeQueueOfClassTreeElems( struct Classtree** queue );
void               PrintQueue( struct Classtree** queue);

/*****************************************************************/

/* classtreereader.c */

#define SIZEOFBUFFER 1000

/*
int               countIndent( char* s);
void              removeNewline( char * s);
*/
struct Classtree* GetClasses( const char * extension);

/*****************************************************************/

int               NumberClassTree( struct Classtree* tree, int from);
int               NumberDescendantsClassTree(struct Classtree* tree);
void              ConstructInvertedIndex(struct Classtree* tree, 
					 struct Classtree**InvertedNodeIndex);
int               GetIndexFromNode( struct Classtree* tree);
/*struct Classtree* GetNodeFromIndex2( int index, struct Classtree* tree);*/
struct Classtree* GetNodeFromIndex( int index, struct Classtree** InvertedNodeIndex);
struct Classtree* GetNodeFromName( struct Classtree* tree, char* name);
int               GetTreeSize( struct Classtree * tree);
/*int               GetNumDescFromIndex2( int index, struct Classtree* tree);*/
int               GetNumDescFromIndex( int index, struct Classtree** InvertedNodeIndex);

/*****************************************************************/

ClassList NewClassList( int size);
void      FreeClassList( ClassList classlist);
void      CopyClassList( ClassList classlist1, ClassList classlist2, 
                         int length);
void      CopyClassMaskedList( ClassList classlist1, ClassList classlist2, 
                               ClassList mask, int length);
void      ApplyMask( ClassList classlist, ClassList mask, 
                     ClassList resultlist, int length);
void      SetBits( ClassList mask,struct Classtree** ontrees, int masklength);
void      SetBitsOffAndOn( ClassList mask, struct Classtree* offtree, 
                           struct Classtree** ontrees, int masklength);
Boolean   AreEqual(ClassList list1, ClassList list2, ClassList mask, int length);
void      PrintClassList(ClassList list, struct Classtree* tree, int length);
void      PrintMaskedClassList(ClassList list, ClassList mask,struct Classtree* tree, int length);

int       NumberOfClassListClasses(ClassList pattern,ClassList mask, int length);
