/*************************************************************************/
/*                                                                       */
/*    classtreereader.c                                                  */
/*    Functions for reading in a Classtree                               */
/*                                                                       */
/*************************************************************************/
#include <stdio.h> 
#include <stdlib.h>
#include <string.h>
#include "classtree.h"
#include "extern.i"


/*****************************************************************/

/*****************************************************************/

static void removeNewline(char * s)
{

  while(s[0] != '\n')
    { s++; } 
  if(s[0] == '\n') { s[0] = '\0'; }

}

/*****************************************************************/

static int countIndent(char* s)
{
  int i = 0;
  while(*s != '\0' && *s == ' ')
    {
      i++; s++;
    } 
  if (*s == '\0') { return -1; }
  else { return i; }
}

/*****************************************************************/

struct Classtree* GetClasses(const char * extension)
{
  FILE *fp;
  char str[SIZEOFBUFFER];
  int lastIndent = -1;
  int currentIndent = 0;
  int linenumber = 0;
  struct Classtree *root,*parenttobe,*sibtobe,*temp;
  int i; 
  char Fn[100];

  strcpy(Fn,FileName);
  strcat(Fn,extension);

  if((fp = fopen(Fn,"r"))==NULL) {
    printf("Cannot open file %s\n",Fn);
    exit(1);
  }

  root = NewClassTreeElem("root",NULL,NULL,NULL);
  parenttobe = root; sibtobe = NULL;
  while(!feof(fp))
    {
      if(fgets(str,SIZEOFBUFFER-1,fp))
	{
	  linenumber++;
	  /*	  printf("<%s>",str); */
	  currentIndent = countIndent(str);
	  /* printf("Currentindent %d\n",currentIndent); */
	  if(currentIndent == -1) {continue; } /* empty line! */

	  else {
	    removeNewline(str);
	    if(currentIndent == lastIndent + 1 ) {
	      /* printf("new child <%s>\n",str+currentIndent); */
	      temp = NewClassTreeElem(str+currentIndent,NULL,parenttobe,NULL);
	      SetChild(parenttobe,temp);
	      parenttobe = temp;
	      sibtobe = temp;
	    }
	    else if(currentIndent == lastIndent) {
	      /* printf("anotherSibling <%s>\n",str+currentIndent); */
	      temp = NewClassTreeElem(str+currentIndent,NULL,parenttobe->parent,NULL);
	      SetSibling(sibtobe,temp);
	      parenttobe = temp;
	      sibtobe = temp;
	    }
	    else if(currentIndent < lastIndent) {
	      /* printf("back up <%s>\n",str+currentIndent); */
	      /* printf("parent to be %s\n",parenttobe->name); */
	      parenttobe = parenttobe->parent;
	      /* printf("parent of parent to be %s\n",parenttobe->name); */
	      for(i = 1; i <= lastIndent - currentIndent; i++) {
		parenttobe = parenttobe->parent;
		/* printf("real parent %s\n",parenttobe->name); */
	      }
	      temp = NewClassTreeElem(str+currentIndent,NULL,parenttobe,NULL);
	      if(parenttobe->child == NULL) {
		SetChild(parenttobe,temp);
		parenttobe = temp;
		sibtobe = temp;
	      }
	      else {
		sibtobe = parenttobe->child;
		while(sibtobe->sibling)
		  {
		    sibtobe = sibtobe->sibling;
		  }		
		SetSibling(sibtobe,temp);
		parenttobe = temp;
		sibtobe = temp;
	      }
	    }
	    else {
	      fprintf(stderr,"Parse error in indentation in line %d (too much indentation)\n",linenumber);
	      exit(1);
	    }
	  }
	}
      lastIndent = currentIndent;
    }
  NumberClassTree(root,0); /* ignoring return value, whwich is size of tree */
  NumberDescendantsClassTree(root);

  return root;
}




