/*************************************************************************/
/*                                                                	 */
/*	Evaluation of a test on a continuous valued attribute	  	 */
/*	-----------------------------------------------------	  	 */
/*								  	 */
/*************************************************************************/


#include "buildex.i"


float
	*SplitGain,	/* SplitGain[i] = gain with att value of item i as threshold */
	*SplitInfo;	/* SplitInfo[i] = potential info ditto */



/*************************************************************************/
/*								  	 */
/*  Continuous attributes are treated as if they have possible values	 */
/*	0 (unknown), 1 (less than cut), 2(greater than cut)	  	 */
/*  This routine finds the best cut for items Fp through Lp and sets	 */
/*  Info[], Gain[] and Bar[]						 */
/*								  	 */
/*************************************************************************/


    EvalContinuousAtt(Att, Fp, Lp)
/*  -----------------  */ 
    Attribute Att;
    ItemNo Fp, Lp; 
{ 
    ItemNo i, BestI, Xp, Tries=0;
    ItemCount Items, KnownItems, LowItems, MinSplit, CountItems();
    ClassNo c;
    float AvGain=0, Val, BestVal, BaseInfo, ThreshCost;
    float ComputeGain(), TotalInfo(), Worth(), MyMaskedInfo(), MySplitInfo();
    void Swap(), ResetFreq();

    short n; 

    ClassList tempclassmask,classmask, classes, existingclasses, ClassesOf();
    struct Classtree **queue, **children, *class;
    int length = MaxClass+1;
    
    float tempgain = -Epsilon, finalgain = -Epsilon;
    float tempinfo = 0.0, finalinfo = 0.0;
    float tempbar  = 0.0, finalbar  = 0.0;


    /*printf("eval contin\n"); *//*ajc99*/

    Verbosity(2) printf("\tAtt %s", AttName[Att]);
    Verbosity(3) printf("\n");

    ResetFreq(2);

    /*  Omit and count unknown values */

    Items = CountItems(Fp, Lp);
    Xp = Fp;
    ValFreq[0] = 0;
    ForEach(i, Fp, Lp)
    {
	if ( CVal(Item[i],Att) == Unknown )
	{
	  /* we don't use Freq[0] at all as far as I can see! 
	     Freq[ 0 ][ ClassAt(Item[i],1) ] += Weight[i]; */
	  Swap(Xp, i);
	  Xp++;
	  ValFreq[0] += Weight[i];
	}
    }
    KnownItems = Items - ValFreq[0];
    UnknownRate[Att] = 1.0 - KnownItems / Items;

    /*  Special case when very few known values  */

    if ( KnownItems < 2 * MINOBJS )
    {
	Verbosity(2) printf("\tinsufficient cases with known values\n");

	Gain[Att] = -Epsilon;
	Info[Att] = 0.0;
	return;
    }

    Quicksort(Xp, Lp, Att, Swap);


    queue           = NewQueueOfClassTreeElems(length);
    children        = NewQueueOfClassTreeElems2(length);
    classmask       = NewClassList(length);
    tempclassmask   = NewClassList(length);
    existingclasses = NewClassList(length);

    /* Find out what classes we really have in this data */
    ForEach(i, Fp, Lp)
      {
	classes = ClassesOf(Item[i]);
	ForEach(n,0,MaxClass)
	  {
	     if(classes[n]) 
	      { existingclasses[n] = true; }
	  }
      }

    /* start queue here */

    /*printf("start queue\n"); *//*ajc99*/
    AddOneToQueue(queue,classTree);

    while(!EmptyQueue(queue))
      {
	/* printf("consider children\n"); *//*ajc99*/
	class = PopFromQueue( queue );
	if(GetListOfChildrenIfExist( class, children, existingclasses ))
	  {
	    CopyClassList( tempclassmask, classmask, length );
	    SetBitsOffAndOn( tempclassmask, class, children, length);
	    /*	    printf("get base info\n"); *//*ajc99*/ 

	    /*  Count base values and determine base information  */

	    ForEach(i, Xp, Lp)
	      {
		classes = ClassesOf(Item[i]);
		ForEach(n,0,MaxClass)
		  { if (classes[n] && tempclassmask[n]) 
		    { Freq[ 2 ][ n ] += Weight[i]; }
		  }
		SplitGain[i] = -Epsilon;
		SplitInfo[i] = 0;
	      }

	    BaseInfo = MyMaskedInfo(Freq[2],0,MaxClass,tempclassmask);

    /*  Try possible cuts between items i and i+1, and determine the
	information and gain of the split in each case.  We have to be wary
	of splitting a small number of items off one end, as we can always
	split off a single item, but this has little predictive power.  */

	    MinSplit = 0.10 * KnownItems / (MaxClass + 1);
	    if ( MinSplit <= MINOBJS ) MinSplit = MINOBJS;
	    else
	      if ( MinSplit > 25 ) MinSplit = 25;
	    
	    /* where is AvGain used? */

	    /*printf("try possible cuts\n"); *//*ajc99*/

	    LowItems = 0; Tries = 0; AvGain = 0;
	    ForEach(i, Xp, Lp - 1)
	      {
		classes = ClassesOf(Item[i]);
		ForEach(c,0,MaxClass)
		  {
		    if(classes[c] && tempclassmask[c])
		      {
			Freq[1][c] += Weight[i];
			Freq[2][c] -= Weight[i];
		      }
		  }
		LowItems   += Weight[i];
	  
		if ( LowItems < MinSplit ) continue;
		else
		  if ( LowItems > KnownItems - MinSplit ) break;
		
		if ( CVal(Item[i],Att) < CVal(Item[i+1],Att) - 1E-5 )
		  {
		    ValFreq[1] = LowItems;
		    ValFreq[2] = KnownItems - LowItems;
		    SplitGain[i] = ComputeGain(BaseInfo, UnknownRate[Att], 2, KnownItems,tempclassmask);
		    SplitInfo[i] = MySplitInfo(ValFreq,1,2); 
		    AvGain += SplitGain[i];
		    Tries++;

		    Verbosity(3)
		      {	printf("\t\tCut at %.3f  (gain %.3f, val %.3f):",
			     ( CVal(Item[i],Att) + CVal(Item[i+1],Att) ) / 2,
			       SplitGain[i],
			       Worth(SplitInfo[i], SplitGain[i], Epsilon));
		        PrintDistribution(Att, 2, true);
		      }
		  }
	      }

	    /*  Find the best attribute according to the given criterion  */

	    /*printf("find best att\n"); *//*ajc99*/
	    ThreshCost = Log(Tries) / Items;

	    BestVal = 0;
	    BestI   = None;
	    ForEach(i, Xp, Lp - 1)
	      {
		if ( (Val = SplitGain[i] - ThreshCost) > BestVal )
		  {
		    BestI   = i;
		    BestVal = Val;
		  }
	      }

	    /*  If a test on the attribute is able to make a gain,
		set the best break point, gain and information  */ 

	    if ( BestI == None )
	      {
		tempgain = -Epsilon;
		tempinfo = 0.0;
		Verbosity(2) printf("\tno gain\n");
	      }
	    else
	      {
		tempbar = (CVal(Item[BestI],Att) + CVal(Item[BestI+1],Att)) / 2;
		tempgain = BestVal;
		tempinfo = SplitInfo[BestI];
		Verbosity(2)
		  printf("\tcut=%.3f, inf %.3f, gain %.3f\n",
			 tempbar,tempinfo,tempgain);
		/*       Bar[Att], Info[Att], Gain[Att]); */
	      }
	  }

	/*printf("keep this one\n"); *//*ajc99*/
	if(tempgain > finalgain)
	  {
	    finalgain = tempgain;
	    finalbar  = tempbar;
	    finalinfo = tempinfo;
	    CopyClassList(classmask,tempclassmask,length);
	    AddToQueue(queue,children);
	  } 
      }
    Gain[Att] = finalgain;
    Bar[Att]  = finalbar;
    Info[Att] = finalinfo;

    FreeClassList(classmask);
    FreeClassList(tempclassmask);
    FreeClassList(existingclasses);
    FreeQueueOfClassTreeElems(queue);
    FreeQueueOfClassTreeElems(children);
}



/*************************************************************************/
/*                                                                	 */
/*  Change a leaf into a test on a continuous attribute           	 */
/*                                                                	 */
/*************************************************************************/


    ContinTest(Node, Att)
/*  ----------  */
    Tree Node;
    Attribute Att;
{
    float Thresh, GreatestValueBelow();
    ItemCount CountItems();

    Sprout(Node, 2);

    Thresh = GreatestValueBelow(Att, Bar[Att]);

    Node->NodeType	= ThreshContin;
    Node->Tested	= Att;
    Node->Cut		=
    Node->Lower		=
    Node->Upper		= Thresh;
    Node->Errors        = 0;
}



/*************************************************************************/
/*                                                                	 */
/*  Return the greatest value of attribute Att below threshold t  	 */
/*                                                                	 */
/*************************************************************************/
 

float GreatestValueBelow(Att, t)
/*    ------------------  */
    Attribute Att;
    float t;
{
    ItemNo i;
    float v, Best;
    Boolean NotYet=true;

    ForEach(i, 0, MaxItem)
    {
	v = CVal(Item[i], Att);
	if ( v != Unknown && v <= t && ( NotYet || v > Best ) )
	{
	    Best = v;
	    NotYet = false;
	}
    }

    return Best;
}
