/*************************************************************************/
/*									 */
/*	Prune a decision tree and predict its error rate		 */
/*	------------------------------------------------		 */
/*									 */
/*************************************************************************/


#include "defns.i"
#include "class.h"  /* contains #include "types.i" */
#include "extern.i"


extern	ItemCount	*Weight;

Set	*PossibleValues=Nil;
Boolean	Changed;

#define	LocalVerbosity(x)	if (Sh >= 0 && VERBOSITY >= x)
#define	Intab(x)		Indent(x, "| ")



/*************************************************************************/
/*									 */
/*  Prune tree T, returning true if tree has been modified		 */
/*									 */
/*************************************************************************/


Boolean Prune(T)
/*      -----  */
    Tree T;
{
    ItemNo i;
    float EstimateErrors();
    void CheckPossibleValues();
    Attribute a;

    printf("Beginning to prune\n"); /*ajc99*/

    InitialiseWeights();
    AllKnown = true;


    Verbosity(1) printf("\n");

    Changed = false;

    EstimateErrors(T, 0, MaxItem, 0, true);

    if ( SUBSET )
    {
	if ( ! PossibleValues )
	{
	    PossibleValues = (Set *) calloc(MaxAtt+1, sizeof(Set));
	}

	ForEach(a, 0, MaxAtt)
	{
	    if ( MaxAttVal[a] )
	    {
		PossibleValues[a] = (Set) malloc((MaxAttVal[a]>>3) + 1);
		ClearBits((MaxAttVal[a]>>3) + 1, PossibleValues[a]);
		ForEach(i, 1, MaxAttVal[a])
		{
		    SetBit(i, PossibleValues[a]);
		}
	    }
	}

	CheckPossibleValues(T);
    }

    printf("Finished pruning\n"); /*ajc99*/

    return Changed;
}




/*************************************************************************/
/*									 */
/*	Estimate the errors in a given subtree				 */
/*									 */
/*************************************************************************/


float EstimateErrors(T, Fp, Lp, Sh, UpdateTree)
/*    --------------  */
    Tree T;
    ItemNo Fp, Lp; 
    short Sh;
    Boolean UpdateTree;
{ 
    ItemNo i, Kp, Ep, Group();
    ItemCount Cases, KnownCases, *LocalClassDist, TreeErrors, LeafErrors,
	ExtraLeafErrors, BranchErrors, CountItems(), Factor, MaxFactor, 
        AddErrs();
    DiscrValue v, MaxBr;
    Boolean PrevAllKnown;
    short n,m;

    short bestpatternsize = 0;
    float bestpatternfreq = 0;

    ClassNo *p;
    int length = MaxClass+1;
    struct Classtree ** queue;

    ClassList classmask, bestpattern;
    ClassList classes, existingclasses;

    float FindBestPattern();
    void RefineBestPattern();

    classmask = NewClassList(length);
    bestpattern = NewClassList(length);
    existingclasses = NewClassList(length);

    /*  Generate the class frequency distribution  */

    Cases = CountItems(Fp,Lp);
    LocalClassDist = (ItemCount *) calloc(MaxClass+1,sizeof(ItemCount));
    
    ForEach(i, Fp, Lp)
      { 
	classes = ClassesOf(Item[i]);
	ForEach(n, 0, MaxClass) {
	  if(classes[n]) { 
	    LocalClassDist[ n ] += Weight[i];
	    existingclasses[n]   = true;
	  } 
	}
      }

    
    /*  Find the most frequent class and update the tree  */
    queue = NewQueueOfClassTreeElems(MaxClass+1);

    if(!GetListOfChildrenIfExist(classTree,queue,existingclasses)) 
      { printf("prune: Nothing in classtree!\n"); exit(1); }
    SetBits(classmask,queue,MaxClass+1);

    bestpatternfreq = FindBestPattern( classmask, bestpattern, 
                            bestpatternfreq, Fp, Lp);
    RefineBestPattern(classmask, bestpattern, bestpatternfreq, queue, Fp,Lp, existingclasses);
    FreeQueueOfClassTreeElems(queue);
    FreeClassList(existingclasses);

    LeafErrors = Cases - bestpatternfreq;
    ExtraLeafErrors = AddErrs(Cases, LeafErrors);

    /* Find the number of classes in the pattern */
    bestpatternsize=0; 
    ForEach (i, 0, MaxClass)
      {
	if(classmask[i] && bestpattern[i]) { 
	  bestpatternsize++;
	}
      }
    
    if ( UpdateTree )
    {
	T->Items = Cases;
	p = (ClassNo *) calloc(bestpatternsize,sizeof(ClassNo));
	/*	ForEach(m,1,bestpatternsize)
	  {
	    p[m-1] = bestpattern[m-1];
	  }
	*/
	i = 0;
	ForEach (m, 0, MaxClass)
	  {
	    if(classmask[m] && bestpattern[m]) { 
	      p[i] = m; i++;
	    }
	  }

	cfree(T->Leaf);
	T->Leaf  = p;
	T->LeafCount = bestpatternsize;
	memcpy(T->ClassDist, LocalClassDist, 
                                    (MaxClass + 1) * sizeof(ItemCount));
    }

    if ( ! T->NodeType )	/*  leaf  */
    {
	TreeErrors = LeafErrors + ExtraLeafErrors;

	if ( UpdateTree )
	{
	    T->Errors = TreeErrors;

	    LocalVerbosity(1)
	    {
		Intab(Sh);
		ForEach(m,0,T->LeafCount -1)
		  printf("%s ",ClassName[T->Leaf[m]]);
		printf("(%.2f:%.2f/%.2f)\n", T->Items, LeafErrors, T->Errors);
	    }
	}

	free(LocalClassDist);
	FreeClassList(classmask);
	FreeClassList(bestpattern);

	return TreeErrors;
    }

    /*  Estimate errors for each branch  */

    Kp = Group(0, Fp, Lp, T) + 1;
    KnownCases = CountItems(Kp, Lp);

    PrevAllKnown = AllKnown;
    if ( Kp != Fp ) AllKnown = false;

    TreeErrors = MaxFactor = 0;

    ForEach(v, 1, T->Forks)
    {
	Ep = Group(v, Kp, Lp, T);

	if ( Kp <= Ep )
	{
	    Factor = CountItems(Kp, Ep) / KnownCases;

	    if ( Factor >= MaxFactor )
	    {
		MaxBr = v;
		MaxFactor = Factor;
	    }

	    ForEach(i, Fp, Kp-1)
	    {
		Weight[i] *= Factor;
	    }

	    TreeErrors += EstimateErrors(T->Branch[v], Fp, Ep, Sh+1, UpdateTree);

	    Group(0, Fp, Ep, T);
	    ForEach(i, Fp, Kp-1)
	    {
		Weight[i] /= Factor;
	    }
	}
    }
 
    AllKnown = PrevAllKnown;

    if ( ! UpdateTree )
    {
	free(LocalClassDist);
	FreeClassList(classmask);
	FreeClassList(bestpattern);
	return TreeErrors;
    }

    /*  See how the largest branch would fare  */

    BranchErrors = EstimateErrors(T->Branch[MaxBr], Fp, Lp, -1000, false);

    LocalVerbosity(1)
    {
        Intab(Sh);
        printf("%s:  [%d%%  N=%.2f  tree=%.2f  leaf=%.2f+%.2f  br[%d]=%.2f]\n",
		AttName[T->Tested],
		(int) ((TreeErrors * 100) / (T->Items + 0.001)),
		T->Items, TreeErrors, LeafErrors, ExtraLeafErrors,
		MaxBr, BranchErrors);
    }

    /*  See whether tree should be replaced with leaf or largest branch  */

    if ( LeafErrors + ExtraLeafErrors <= BranchErrors + 0.1 &&
	 LeafErrors + ExtraLeafErrors <= TreeErrors + 0.1 )
    {
	LocalVerbosity(1)
	{
	    Intab(Sh);
	    printf("Replaced with leaf ");
	    ForEach(m,0,T->LeafCount-1)
	      printf("%s ", ClassName[T->Leaf[m]]);
	    printf("\n");
	}

	T->NodeType = 0;
	T->Errors = LeafErrors + ExtraLeafErrors;
	Changed = true;
    }
    else
    if ( BranchErrors <= TreeErrors + 0.1 )
    {
	LocalVerbosity(1)
	{
	    Intab(Sh);
	    printf("Replaced with branch %d\n", MaxBr);
	}

	AllKnown = PrevAllKnown;
	EstimateErrors(T->Branch[MaxBr], Fp, Lp, Sh, true);
	memcpy((char *) T, (char *) T->Branch[MaxBr], sizeof(TreeRec));
	Changed = true;
    }
    else
    {
	T->Errors = TreeErrors;
    }

    AllKnown = PrevAllKnown;

    /* free memory */
    free(LocalClassDist);
    FreeClassList(classmask);
    FreeClassList(bestpattern);

    return T->Errors;
}



/*************************************************************************/
/*									 */
/*	Remove unnecessary subset tests on missing values		 */
/*									 */
/*************************************************************************/


void CheckPossibleValues(T)
/*  -------------------  */
    Tree T;
{
    Set HoldValues;
    int v, Bytes, b;
    Attribute A;
    char Any=0;

    if ( T->NodeType == BrSubset )
    {
	A = T->Tested;

	Bytes = (MaxAttVal[A]>>3) + 1;
	HoldValues = (Set) malloc(Bytes);

	/*  See if last (default) branch can be simplified or omitted  */

	ForEach(b, 0, Bytes-1)
	{
	    T->Subset[T->Forks][b] &= PossibleValues[A][b];
	    Any |= T->Subset[T->Forks][b];
	}

	if ( ! Any )
	{
	    T->Forks--;
	}

	/*  Process each subtree, leaving only values in branch subset  */

	CopyBits(Bytes, PossibleValues[A], HoldValues);

	ForEach(v, 1, T->Forks)
	{
	    CopyBits(Bytes, T->Subset[v], PossibleValues[A]);

	    CheckPossibleValues(T->Branch[v]);
	}

	CopyBits(Bytes, HoldValues, PossibleValues[A]);

	free(HoldValues);
    }
    else
    if ( T->NodeType )
    {
	ForEach(v, 1, T->Forks)
	{
	    CheckPossibleValues(T->Branch[v]);
	}
    }
}
